import argparse
import logging
import sys

import psycopg2
from psycopg2 import sql
from psycopg2 import extras

from . import parser

log = logging.getLogger(__name__)
log_action = log.getChild("action")
log_check = log.getChild("check")
log_state = log.getChild("state")

class Node:
    def __init__(self, name, depends):
        self.name = name
        self.depends = depends
        self.ok = False
        self.ready = False

    def __repr__(self):
        return f"{type(self)}({self.name}{' ready' if self.ready else ''}{' ok' if self.ok else ''})"

    def is_ready(self):
        # XXX - Naive O(n²) algorithm
        if self.ready:
            return True
        for d in self.depends:
            found = False
            for w in want:
                if w.name == d:
                    if not w.ok:
                        log_state.info("%s depends on %s which is not yet ok", self.name, d)
                        return False
                    found = True
                    break
            if not found:
                raise RuntimeError(f"Dependency {d} of {self.name} doesn't exist")
        log_state.info("%s is now ready", self.name)
        self.ready = True
        return True

    def set_order(self):
        global order
        order += 1
        self.order = order

class HaveData(Node):
    def __init__(self, name, depends, table, key, extra):
        super().__init__(name, depends)
        self.table = table
        self.key = key
        self.extra = extra

    def check(self):
        log_check.info("Checking %s", self.name)
        csr = db.cursor(cursor_factory=extras.DictCursor)
        key_checks = [
            sql.SQL(" = ").join([ sql.Identifier(x), sql.Placeholder() ])
            for x in self.key.keys()
        ]
        key_check = sql.SQL(" and ").join(key_checks)
        q = sql.SQL(
                "select * from {table} where {key_check}"
            ).format(
                table=sql.Identifier(self.table),
                key_check=key_check
            )
        key_values = [v.resolve() if isinstance(v, Ref) else v for v in self.key.values()]
        csr.execute(q, key_values)
        self.result = csr.fetchall()
        log_check.info("Got %d rows", len(self.result))
        if self.result:
            extra_columns = list(self.extra.keys())
            for c in extra_columns:
                if self.result[0][c] != self.extra[c]:
                    log_action.info("Updating %s: %s <- %s", key_values, c, self.extra[c])
                    q = sql.SQL(
                        "update {table} set {column}={placeholder} where {key_check}"
                    ).format(
                        table=sql.Identifier(self.table),
                        column=sql.Identifier(c),
                        placeholder=sql.Placeholder(),
                        key_check=key_check,
                    )
                    csr.execute(q, [self.extra[c]] + key_values)
                    self.result[0][c] = self.extra[c]
            self.set_order()
            self.ok = True
            log_state.info("%s is now ok", self.name)
            return
        else:
            extra_values = [v.resolve() if isinstance(v, Ref) else v for v in self.extra.values()]
            columns = list(self.key.keys()) + list(self.extra.keys())
            values = key_values + extra_values
            q = sql.SQL(
                    "insert into {table}({columns}) values({placeholders}) returning *"
                ).format(
                    table=sql.Identifier(self.table),
                    columns=sql.SQL(", ").join([sql.Identifier(x) for x in columns]),
                    placeholders=sql.SQL(", ").join([sql.Placeholder() for x in columns]),
                )
            log_action.info("Inserting data")
            csr.execute(q, values)
            self.result = csr.fetchall()
            log_action.info("Got %d rows", len(self.result))
            if self.result:
                self.set_order()
                self.ok = True
                log_state.info("%s is now ok", self.name)
                return
            # We shouldn't get here. Either the insert succeeded, or it raised an
            # exception. Success with 0 rows should not happen.
        raise RuntimeError("Unreachable code reached")

class HaveInit(Node):
    def __init__(self, name, depends, table, query):
        super().__init__(name, depends)
        self.table = table
        self.query = query

    def check(self):
        log_check.info("Checking %s", self.name)

        for w in want:
            if isinstance(w, HaveTable) and w.table == self.table:
                if not w.ok:
                    log_check.error("%s not yet ok", w.name)
                    raise RuntimeError(f"Cannot insert into table {w.table} from {w.name} which is not yet ok. Please add a dependency")
                if w.new:
                    log_action.info("Executing %s", self.query)
                    csr = db.cursor(cursor_factory=extras.DictCursor)
                    csr.execute(self.query)
                    self.result = csr.fetchall()
                    log_action.info("Got %d rows", len(self.result))
                    self.set_order()
                else:
                    log_check.info("Table %s already exists", w.table)
                self.ok = True
                log_state.info("%s is now ok", self.name)
                break
        else:
            raise RuntimeError(f"Cannot find a rule which creates table {self.table}")


class HaveTable(Node):

    def __init__(self, name, depends, table, schema="public"):
        super().__init__(name, depends)
        self.table = table
        self.schema = "public"

    def check(self):
        log_check.info("Checking %s", self.name)
        csr = db.cursor(cursor_factory=extras.DictCursor)
        csr.execute(
                """
                select * from information_schema.tables
                where table_schema = %s and table_name = %s
                """,
                (self.schema, self.table,))
        r = csr.fetchall()
        if len(r) == 1:
            # Table exists, all ok
            self.set_order()
            self.ok = True
            self.new = False
            log_state.info("%s is now ok", self.name)
            return
        if len(r) > 1:
            raise RuntimeError(f"Found {len(r)} tables with schema {self.schema} and name {self.table}")

        # Create table
        # (Yes, we can actually create a table with 0 columns)
        log_action.info("Creating table %s.%s", self.schema, self.table)
        q = sql.SQL("create table {schema}.{table}()").format(
                schema=sql.Identifier(self.schema),
                table=sql.Identifier(self.table)
            )
        csr.execute(q)
        self.set_order()
        self.ok = True
        self.new = True
        log_state.info("%s is now ok", self.name)

    pass

class HaveColumn(Node):
    # hjp=> alter table service add id serial primary key;
    # ALTER TABLE
    # hjp=> alter table service add type text;
    # ALTER TABLE
    # ...
    def __init__(self, name, depends, table, column, definition, schema="public"):
        super().__init__(name, depends)
        self.table = table
        self.column = column
        self.definition = definition
        self.schema = "public"

    def check(self):
        log_check.info("Checking %s", self.name)
        csr = db.cursor(cursor_factory=extras.DictCursor)
        # For now just check if column exists. Checking the type etc. will be implemented later
        csr.execute(
                """
                select * from information_schema.columns
                where table_schema = %s and table_name = %s and column_name = %s
                """,
                (self.schema, self.table, self.column, ))
        r = csr.fetchall()
        if len(r) == 1:
            # Column exists, all ok
            self.set_order()
            self.ok = True
            log_state.info("%s is now ok", self.name)
            return
        if len(r) > 1:
            raise RuntimeError(f"Found {len(r)} columns with nam {self.columnr} in {self.schema}.{self.table}")

        # Create column
        log_action.info("Adding column %s to table %s.%s", self.column, self.schema, self.table)
        q = sql.SQL("alter table {schema}.{table} add {column} {definition}").format(
                schema=sql.Identifier(self.schema),
                table=sql.Identifier(self.table),
                column=sql.Identifier(self.column),
                definition=sql.SQL(self.definition),
            )
        csr.execute(q)
        self.set_order()
        self.ok = True
        log_state.info("%s is now ok", self.name)

class HaveUniqueConstraint(Node):
    # hjp=> alter table service add unique (type, feature);
    # ALTER TABLE
    pass

class HaveIndex(Node):
    def __init__(self, name, depends, table, index, type, definition, schema="public"):
        super().__init__(name, depends)
        self.table = table
        self.index = index
        self.type = type
        self.definition = definition
        self.schema = schema

    def check(self):
        log_check.info("Checking %s", self.name)
        csr = db.cursor(cursor_factory=extras.DictCursor)
        # For now just check if index exists. Checking the type etc. will be implemented later
        csr.execute(
                """
                select * from pg_indexes
                where schemaname = %s and tablename = %s and indexname = %s
                """,
                (self.schema, self.table, self.index, ))
        r = csr.fetchall()
        if len(r) == 1:
            # Index exists, all ok
            self.set_order()
            self.ok = True
            log_state.info("%s is now ok", self.name)
            return
        if len(r) > 1:
            raise RuntimeError(f"Found {len(r)} indexes with name {self.index} on {self.schema}.{self.table}")

        # Create index
        log_action.info("Adding index %s to table %s.%s", self.index, self.schema, self.table)
        q = sql.SQL("create {type} {index} on {schema}.{table} {definition}").format(
                type=sql.SQL(self.type),
                schema=sql.Identifier(self.schema),
                table=sql.Identifier(self.table),
                index=sql.Identifier(self.index),
                definition=sql.SQL(self.definition),
            )
        csr.execute(q)
        self.set_order()
        self.ok = True
        log_state.info("%s is now ok", self.name)


def findnode(name):
    for w in want:
        if w.name == name:
            return w

class Ref:
    def __init__(self, datanode, row, column):
        self.datanode = datanode
        self.row = row
        self.column = column

    def resolve(self):
        datanode = findnode(self.datanode)
        if not datanode.ok:
            # XXX - We might try to resolve this, but for now the user is responsible to declare the dependency explicitely
            raise RuntimeError(f"Cannot get data from {datanode.name} which is not yet ok. Please add a dependency")
        return datanode.result[self.row][self.column]


def fit(_db, _want):
    global db, want, order
    db = _db
    want = _want
    order = 0
    while True:
        progress = False
        not_ok = 0
        for w in want:
            if not w.ok:
                if w.is_ready():
                    w.check()
                    progress = True
                else:
                    not_ok += 1
        if not_ok == 0:
            break
        if not progress:
            raise RuntimeError(f"Didn't make any progress in this round, but {not_ok} requirements are still not ok")
        log_state.info("%d requirements are not yet ok", not_ok)

    db.commit()
    log_state.info("Done")

def dump_dot():
    print("digraph {")
    for w in want:
        print(f"\t{w.name} [shape=rect]")
        for d in w.depends:
            print(f"\t{d} -> {w.name} [constraint=false]")
    in_order = [w.name for w in sorted(want, key=lambda x: x.order)]
    for i in range(1, len(in_order)):
        print(f"\t{in_order[i-1]} -> {in_order[i]} [style=dashed]")
    print("}")

def main():
    logging.basicConfig(format="%(asctime)s %(levelname)s %(name)s %(lineno)d | %(message)s", level=logging.DEBUG)
    ap = argparse.ArgumentParser()
    ap.add_argument("--dbname")
    ap.add_argument("--dbuser")
    ap.add_argument("files", nargs="+")
    args = ap.parse_args()

    db = psycopg2.connect(dbname=args.dbname, user=args.dbuser)

    rules = []
    for f in args.files:
        with open(f) as rf:
            text = rf.read()
        ps = parser.ParseState(text)

        ps2 = parser.parse_ruleset(ps)

        if not ps2:
            ps.printerror()
            sys.exit(1)
        rules.extend(ps2.ast)

    fit(db, rules)




#!/usr/bin/python3
import logging
import re
import sys
from pprint import pprint

import psycopg2

import procrusql

class RuleFile:
    def __init__(self, filename):
        with open(filename) as f:
            self.text = f.read()

class Failure:
    def __init__(self, message, position, parse_state):
        self.message = message
        self.position = position
        self.parse_state = parse_state

class ParseState:

    def __init__(self, text, position=0):
        self.text = text
        self.position = position
        self.child_failure = None

    def clone(self):
        ps = ParseState(self.text, self.position)
        return ps

    @property
    def rest(self):
        return self.text[self.position:]

    def printerror(self):
        if not self.child_failure:
            return
        position = self.child_failure.position
        message = self.child_failure.message
        linesbefore = self.text[:position].split("\n")
        linesafter = self.text[position:].split("\n")
        good = "\x1B[40;32m"
        bad = "\x1B[40;31m"
        reset = "\x1B[0m"
        s = reset + message + "\n"
        lines = []
        for ln in range(max(len(linesbefore) - 3, 0), len(linesbefore)):
            # XXX - join
            lines.append(reset + f"{ln+1:4}: " + good + linesbefore[ln] + reset)
        s += "\n".join(lines)
        s += bad + linesafter[0] + reset
        print(s)

        if  self.child_failure.parse_state:
            self.child_failure.parse_state.printerror()

    def skip_whitespace_and_comments(self):
        self.match(r"(\s+|#.*)*")

    def match(self, regexp):
        if m := re.match(regexp, self.text[self.position:]):
            self.position += len(m.group(0))
        return m

    def match_newlines(self):
        self.match(r"(\s+|#.*)*\n")

    def record_child_failure(self, ps_child, msg):
        if not self.child_failure or  ps_child.position > self.child_failure.position:
            self.child_failure = Failure(position=ps_child.position, message=msg, parse_state=ps_child)

def parse_ruleset(ps):
    ps2 = ps.clone()
    ps2.ast = []
    while ps2.rest:
        ps3 = parse_table_rule(ps2) or  \
              parse_column_rule(ps2) or \
              parse_data_rule(ps2) or   \
              parse_index_rule(ps2)
        if ps3:
            ps2.ast.append(ps3.ast)
            ps2.position = ps3.position
        else:
            ps.record_child_failure(ps2, "expected one of: table rule, column rule, data rule")
            return
    return ps2

def parse_table_rule(ps):
    ps2 = ps.clone()
    ps2.skip_whitespace_and_comments()
    if not ps2.match(r"table\b"):
        ps.record_child_failure(ps2, "expected “table”")
        return
    ps2.skip_whitespace_and_comments()
    ps3 = parse_table_name(ps2)
    if not ps3:
        ps.record_child_failure(ps2, "expected table name")
        return

    ps2.ast = procrusql.HaveTable(rulename(), [], ps3.ast[0])
    ps2.position = ps3.position
    return ps2

def parse_column_rule(ps):
    ps2 = ps.clone()
    ps2.ast = []
    ps2.skip_whitespace_and_comments()
    if not ps2.match(r"column\b"):
        ps.record_child_failure(ps2, "expected “column”")
        return

    # The table name should be omitted if this is part of a table declaration.
    # I haven't decided if I want to make that optional in this rule or write a
    # different rule. Probably the latter. If the former, I may have to change
    # the syntax to avoid ambiguity.
    ps3 = parse_table_name(ps2)
    if not ps3:
        ps.record_child_failure(ps2, "expected table name")
        return
    table_name = ps3.ast[0]
    ps2.position = ps3.position

    ps3 = parse_column_name(ps2)
    if not ps3:
        ps.record_child_failure(ps2, "expected column name")
        return
    column_name = ps3.ast[0]
    ps2.position = ps3.position

    ps3 = parse_column_definition(ps2)
    if not ps3:
        ps.record_child_failure(ps2, "expected column definition")
        return
    column_definition = ps3.ast[0]
    ps2.position = ps3.position

    ps2.ast = procrusql.HaveColumn(rulename(), [], table_name, column_name, column_definition)

    ps2.match_newlines()

    return ps2

def parse_data_rule(ps):
    ps2 = ps.clone()
    ps2.skip_whitespace_and_comments()
    if not ps2.match(r"data\b"):
        ps.record_child_failure(ps2, "expected “data”")
        return

    ps3 = parse_table_name(ps2)
    if not ps3:
        ps.record_child_failure(ps2, "expected table name")
        return
    table_name = ps3.ast[0]
    ps2.position = ps3.position

    if ps3 := parse_dict(ps2):
        key_data = ps3.ast
        ps2.position = ps3.position

        ps3 = parse_dict(ps2)
        if not ps3:
            ps.record_child_failure(ps2, "expected extra data definition")
            return
        extra_data = ps3.ast
        ps2.position = ps3.position

        ps3 = parse_label(ps2)
        if ps3:
            label = ps3.ast
            ps2.position = ps3.position
        else:
            label = rulename()

        ps2.ast = procrusql.HaveData(label, [], table_name, key_data, extra_data)

        ps2.match_newlines()
    elif ps3 := parse_init_query(ps2):
        # We have a bit of a problem here: The query extends to the end of the
        # line so there is no room for a label. I really don't want to parse
        # SQL here.
        label = rulename()
        ps2.position = ps3.position
        ps2.ast = procrusql.HaveInit(label, [], table_name, ps3.ast)

        ps2.match_newlines()
    else:
        ps.record_child_failure(ps2, "expected key data definition or insert query")
        return

    return ps2

def parse_index_rule(ps):
    ps2 = ps.clone()
    ps2.skip_whitespace_and_comments()
    m = ps2.match(r"(unique)? index\b")
    if not m:
        ps.record_child_failure(ps2, "expected “(unique) index”")
        return
    index_type = m.group(0)

    ps3 = parse_index_name(ps2)
    if not ps3:
        ps.record_child_failure(ps2, "expected index name")
        return
    index_name = ps3.ast[0]
    ps2.position = ps3.position
    if not ps2.match(r"\s+on\b"):
        ps.record_child_failure(ps2, "expected “on”")
        return

    ps3 = parse_table_name(ps2)
    if not ps3:
        ps.record_child_failure(ps2, "expected table name")
        return
    table_name = ps3.ast[0]
    ps2.position = ps3.position

    m = ps2.match(r"\s*(using\b|\([\w, ]+\))[^>\n]*")
    if not m:
        ps.record_child_failure(ps2, "expected “using” or column list")
    index_definition = m.group(0)

    ps3 = parse_label(ps2)
    if ps3:
        label = ps3.ast
        ps2.position = ps3.position
    else:
        label = rulename()

    ps2.ast = procrusql.HaveIndex(label, [], table_name, index_name, index_type, index_definition)

    ps2.match_newlines()

    return ps2



def parse_table_name(ps):
    # For now this matches only simple names, not schema-qualified names or
    # quoted names.
    ps2 = ps.clone()
    ps2.ast = []
    ps2.skip_whitespace_and_comments()
    if ps2.rest[0].isalpha():
        m = ps2.match(r"\w+") # always succeeds since we already checked the first character
        ps2.ast.append(m.group(0))
    else:
        ps.record_child_failure(ps2, "expected table name")
    return ps2

def parse_column_name(ps):
    # For now this matches only simple names, not quoted names.
    # Also, this is an exact duplicate of parse_table_name, but they will
    # probably diverge, so I duplicated it.
    ps2 = ps.clone()
    ps2.ast = []
    ps2.skip_whitespace_and_comments()
    if ps2.rest[0].isalpha():
        m = ps2.match(r"\w+") # always succeeds since we already checked the first character
        ps2.ast.append(m.group(0))
        return ps2
    else:
        ps.record_child_failure(ps2, "expected column name")
        return

def parse_index_name(ps):
    # this is an exact duplicate of parse_table_name and parse_column_name, but they will
    # probably diverge, so I duplicated it. I probably should define a
    # parse_identifier and redefine them in terms of it.
    ps2 = ps.clone()
    ps2.ast = []
    ps2.skip_whitespace_and_comments()
    if ps2.rest[0].isalpha():
        m = ps2.match(r"\w+") # always succeeds since we already checked the first character
        ps2.ast.append(m.group(0))
        return ps2
    else:
        ps.record_child_failure(ps2, "expected index name")
        return

def parse_column_definition(ps):
    ps2 = ps.clone()
    ps2.ast = []
    ps2.skip_whitespace_and_comments()
    sqltypes = (
        "int", "integer", "serial",
        "boolean",
        "text", "character varying",
        "date", "timestamp with time zone", "timestamptz"
    )
    pattern = "(" + "|".join(sqltypes) + ")" + r"([ \t]+(default .*|not null\b|primary key\b|unique\b|references \w+\b))*"
    m = ps2.match(pattern)
    if not m:
        ps.record_child_failure(ps2, "expected column definition")
        return
    ps2.ast.append(m.group(0))
    return ps2

def parse_dict(ps):
    ps2 = ps.clone()
    d = {}
    ps2.skip_whitespace_and_comments()
    if not ps2.match(r"{"):
        ps.record_child_failure(ps2, "expected “{”")
        return
    while True:
        ps2.skip_whitespace_and_comments()
        if ps2.match(r'}'):
            break

        m = ps2.match(r'\w+|"([^"]+)"')
        if not m:
            ps.record_child_failure(ps2, "expected column name")
            return
        # XXX - unquote properly
        if m.group(1):
            k = m.group(1)
        else:
            k = m.group(0)

        ps2.skip_whitespace_and_comments()
        if not ps2.match(":"):
            ps.record_child_failure(ps2, "expected “:”")
            return
        ps2.skip_whitespace_and_comments()
        if m := ps2.match(r'[0-9]+'):
            v = int(m.group(0))
        elif m := ps2.match(r'"([^"]*)"'):
            # XXX - process backslash escapes
            v = m.group(1)
        elif m := ps2.match(r'[tT]rue'):
            v = True
        elif m := ps2.match(r'[fF]alse'):
            v = False
        elif m := ps2.match(r'None|null|NULL'):
            v = None
        elif m := ps2.match(r'@(\w+)/(\d+)/(\w+)'):
            v = procrusql.Ref(m.group(1), int(m.group(2)), m.group(3))
        else:
            ps.record_child_failure(ps2, "expected value")
            return

        d[k] = v

        ps2.skip_whitespace_and_comments()
        comma_found = ps2.match(r',')
        ps2.skip_whitespace_and_comments()
        if ps2.match(r'}'):
            break
        if not comma_found:
            ps.record_child_failure(ps2, "expected comma or close brace")
            return
    ps2.ast = d
    return ps2

def parse_init_query(ps):
    ps2 = ps.clone()
    ps2.skip_whitespace_and_comments()
    if m := ps2.match(r'(?i:with|insert)\b.*'):
        ps2.ast = m.group(0)
        return ps2
    else:
        ps.record_child_failure(ps2, "expected insert query")
        return

def parse_label(ps):
    ps2 = ps.clone()
    if m := ps2.match(r"\s*>>\s*(\w+)"):
        ps2.ast = m.group(1)
        return ps2
    else:
        ps.record_child_failure(ps2, "expected label definition")
        return

rulenum = 0
def rulename():
    global rulenum
    rulenum += 1
    return f"__rule_{rulenum}"

